/*
 * String2XmString.c - Ein neuer Ressourcen-Konverter, der von String nach
 *                     XmString konvertiert und dabei sowohl Font-Tags als
 *                     auch die Schreibrichtung beruecksichtigen kann.
 *
 * Letzte Aenderung: 20.09.1994 v1.01
 * 
 * (c) 1994 Harald Albrecht
 * albrecht@igpm.rwth-aachen.de
 * file://ftp.igpm.rwth-aachen.de/arc/pub/unix/html/motifcorner.html
 */

#include <Xm/Xm.h>
#include <X11/StringDefs.h>

#if (XmREVISION <= 1) 
#define XmFONTLIST_DEFAULT_TAG XmSTRING_DEFAULT_CHARSET
#endif

/* --------------------------------------------------------------------------
 * Diese Funktion sucht nach dem Ende innerhalb eines Ressourcentexts. Dabei
 * werden die Sonderzeichen "{" und "}" jeweils als Paare beruecksichtigt. 
 * Gesucht wird letzten Endes nach dem abschliessenden "}".
 * 
 * Parameter:
 *   Text	    ^ auf den Text, nach dessen Ende gesucht werden soll.
 * 
 * Ergebnis:
 *   Zeiger auf das erste Zeichen, dass NICHT mehr zum Text gehoert. Das ist
 *   dann entweder ein schliessende geschweifte Klammer "}" oder das Ende des
 *   Strings in Form einer \0.
 */
static char *GetEndOfText(char *Text)
{
    char ch;
    
    int Braces = 1; /* Klammern duerfen nur im Rudel und dann auch nur noch
                       paarweise auftreten! */
    
    for ( ;; ) {
	ch = *Text++;
	if      ( ch == '{' ) ++Braces;
	else if ( ch == '}' ) {
	    if ( --Braces <= 0 ) return Text-1;
	} else if ( ch == 0 ) return Text-1;
    }
} /* GetEndOfText */

/* --------------------------------------------------------------------------
 * Diese Funktion erzeugt eine Kopie eines Stringabschnitts.
 * 
 * Parameter:
 *   Start	^ auf Beginn des Stringabschnitts
 *   End	^ auf das erste nicht mehr mitzukopierende Zeichen oder
 *		aber kann auch NULL sein, wenn bis zum Ende kopiert werden
 *		soll.
 *
 * Ergebnis:
 *   Zeiger auf eine Kopie des angegebenen Stringabschnitts. Nach Gebrauch
 *   bitte wieder mit XtFree entsorgen, ansonsten gibt's irgendwann Speicher-
 *   sorgen!
 */
static char *MakeCopy(char *Start, char *End)
{
    int   size;
    char *Str;
    
    if ( End )
	size = End - Start + 1;
    else
	size = strlen(Start) + 1;
    Str = (char *) XtMalloc(size);
    strncpy(Str, Start, size - 1);
    Str[size - 1] = 0;
    return Str;
} /* MakeCopy */

/* --------------------------------------------------------------------------
 * Erzeugt einen einzelnen XmString mit dem angegebenen FontTag. Zudem werden
 * die guten, alten Zeilenwechsel \n beruecksichtigt und dementsprechend in
 * sog. Separatoren umgewandelt. Der von dieser Funktion generierte String
 * besitzt keine festgelegte Richtung, kann also je nach Kontext von links
 * nach rechts oder aber von rechts nach links laufen.
 *
 * Parameter:
 *   pStr		^ auf einen ganz ordinaeren C-String, Null-terminiert
 *   FontTag		Namen des Font-Tags, welches der zu erzeugende String
 *			zugewiesen bekommen soll.
 *
 * Ergebnis:
 *   Gibt den entsprechenden Motif-String zurueck.
 */
static XmString CreateXmString(char *pStr, char *FontTag)
{
    XmString Str, OldStr, TmpStr, XmSeparator;
    char     *pSeparator;
    
    Str = NULL; XmSeparator = NULL;
    while ( *pStr ) {
    /* 
     * Suche nun jeweils nach dem naechsten Zeilenwechsel innerhalb des
     * Strings. Wird einer gefunden, so wird der String dort durchtrennt,
     * indem wir eine Endemarke \0 dorthin schreiben. In jedem Fall kann
     * dann der String (oder ein Teil davon erzeugt werden.
     */
        pSeparator = strchr(pStr, '\n');
        if ( pSeparator ) *pSeparator = 0;
        TmpStr = XmStringCreate(pStr, FontTag);
        if ( Str ) {
            Str = XmStringConcat(OldStr = Str, TmpStr);
            XmStringFree(OldStr);
            XmStringFree(TmpStr);
        } else
            Str = TmpStr;
    /*
     * Waren wir zuvor auf einen Zeilentrenner \n getreten, dann muessen wir
     * hier noch das Motif-Aequivalent an dem XmString 'dranhaengen.
     */
        if ( pSeparator ) {
            if ( XmSeparator ) TmpStr = XmSeparator;
            else               TmpStr = XmSeparator = XmStringSeparatorCreate();
            Str = XmStringConcat(OldStr = Str, TmpStr);
            XmStringFree(OldStr);
            pStr = pSeparator + 1;
        } else
            break; /* Wir sind damit dann auch fertig! */
    }
    /* Nach der Arbeit wieder evtl. Altlasten entsorgen! */
    if ( XmSeparator ) XmStringFree(XmSeparator);
    return Str;
} /* CreateXmString */

/* --------------------------------------------------------------------------
 * Hier wird ein String (bzw. ein Teil davon, dann zeigt stringEnd auf das
 * erste nicht mehr zu beruecksichtigende Zeichen) in einen ordentlichen
 * XmString umwandeln. Dabei koennen im Gegensatz zu der Magerkost aus der
 * OSF-Kueche auch Compound-Strings erzeugt werden, die unterschiedliche
 * Zeichensaetze resp. Schriftschnitte verwenden. Ausserdem kann auch noch
 * die Schriftrichtung gewechselt werden! Also alles das, was die von der OSF
 * 'mal wieder vergessen haben. Wie ich bereits bei der ComboBox bemerkte:
 * MUSS MAN DENN ALLES SELBST MACHEN?!
 *
 * Parameter:
 *   string		Der String selbiger welcher in Text zusammengesetzter-
 *			weis gewandelt werden soll.
 *   stringEnd		^ auf dasjenige Zeichen, welches nicht mehr zum Text
 *			gehoeren soll oder aber NULL. Zeigt im Normalfall auf
 *			eine schliessende Klasser "}" oder auf eine \0.
 *
 * Ergebnis:
 *   Der entsprechende XmString, ggf. mit unterschiedlichen Font-Tags und
 *   Schriftrichtungen.
 */
static XmString StringToXmString(char *string, char *stringEnd)
{
    XmString Message, tmpStr, OldMessage;
    char     *pStr, *pEnd, *pFontTag;
    char     ch;
    
    string = MakeCopy(string, stringEnd);
    Message = NULL;
    pStr    = string;
    
    if ( *pStr == '\\' ) {
    /* 
     * Nach einem Fluchtsymbol ist das naechste Zeichen literal und es
     * kommt deshalb nur ein stinknormaler Hilfetext, der unabhaengig von
     * der aktuellen Hilfestufe immer angezeigt wird. Wir koennen daher
     * direkt eine Kopie des Strings zurueckgeben, allerdings ohne das
     * Fluchtsymbol am Textbeginn.
     */
	++pStr;
	Message = CreateXmString(pStr, XmFONTLIST_DEFAULT_TAG);
	XtFree(string);
	return Message;
    }
    
    while ( *pStr ) {
    /*
     * Der String kann sich aus mehreren Eintraegen zusammensetzen, die
     * jeweils durch { und } sowie dem dazwischenstehenden Text gebildet
     * werden. Jeder Eintrag kann Informationen ueber FontTag sowie die
     * Schriftrichtung enthalten. ACHTUNG: Bei der Angabe einer Schrift-
     * richtung bleibt diese Richtung auch in den folgenden Eintraegen
     * solange erhalten, bis sie erneut ueberschrieben wird!
     */
        if ( *pStr == '{' ) {
    /*
     * Zuerst einmal schauen wir nach, ob eine bestimmte Schriftrichtung
     * vorgegeben wird. In diesem Fall muessen wir einen entsprechenden
     * Eintrag an den String anhaengen.
     */
            ch = *++pStr;
            if ( (ch == '>') || (ch == '<') ) {
                tmpStr = XmStringDirectionCreate( ch == '>' ?
                           XmSTRING_DIRECTION_L_TO_R :
                           XmSTRING_DIRECTION_R_TO_L);
                if ( Message ) {
                    Message = XmStringConcat(OldMessage = Message, tmpStr);
                    XmStringFree(OldMessage);
                    XmStringFree(tmpStr);
                } else
                    Message = tmpStr;
                ch = *++pStr; /* weiter mit dem naechsten Zeichen */
            }
    /*
     * Jetzt sollte normalerweise erst einmal der FontTag kommen. Hier ist
     * es aber auch erlaubt, nichts anzugeben und nach einem Leerfeld mit
     * dem eigentlichen Text fortzufahren oder aber ein * anzugeben. Beide
     * Faelle bedeuten uebereinstimmend, dass der Standard-Font-Tag zu
     * benutzen ist.
     */
            pFontTag = pStr;
            while ( ch && (ch != ' ') && (ch != '\t') && (ch != '\n') )
                ch = *++pStr;
            if ( !ch ) {
		XtWarning("malformed help text entry");
                XtFree(string);
                return Message;
            }
            *pStr++ = 0;
            pEnd = GetEndOfText(pStr);
            ch = *pEnd; *pEnd = 0;
            if ( (*pFontTag == ' ') || (*pFontTag == '*') )
                pFontTag = XmFONTLIST_DEFAULT_TAG;
            tmpStr = CreateXmString(pStr, pFontTag);
            if ( Message ) {
                Message = XmStringConcat(OldMessage = Message, tmpStr);
                XmStringFree(OldMessage);
                XmStringFree(tmpStr);
            } else
                Message = tmpStr;
            if ( !ch ) {
                XtFree(string);
                return Message;
            }
            pStr = pEnd + 1;
        } else {
    /*
     * Es folgt nur noch normaler Text, der nicht in {} steht und damit
     * kein explizites FontTag zugewiesen bekommt. Damit sind wir dann
     * auch schon fast fertig.
     */
            if ( Message ) {
                tmpStr = CreateXmString(pStr, XmFONTLIST_DEFAULT_TAG);
                Message = XmStringConcat(OldMessage = Message, tmpStr);
                XmStringFree(OldMessage);
                XmStringFree(tmpStr);
                XtFree(string);
                return Message;
            } else {
                Message = CreateXmString(pStr, XmFONTLIST_DEFAULT_TAG);
                XtFree(string);
                return Message;
            }
        }
    }
    XtFree(string);
    return Message;
} /* StringToXmString */

/* --------------------------------------------------------------------------
 * Jetzt kommen wir so langsam zum eigentlichen Konverter. Die folgenden
 * Routinen bilden die Schnittstelle zwischen obigen Funktionen und dem
 * Toolkit, welches die Konverter verwaltet.
 */
static void CleanupXmString(XtAppContext AppCtx, XrmValue *To, 
                            XtPointer ConverterData,
                            XrmValue Args, Cardinal *NumArgs)
{
    XmStringFree(*(XmString *) To->addr);
} /* CleanupXmString */

static Boolean CvtStringToXmString(Display *Dsp,
                                   XrmValue *Args, Cardinal *NumArgs,
                                   XrmValue *From, XrmValue *To,
                                   XtPointer *ConverterData)
{
    static XmString string; /* !! */
    
    /*
     * Ein nicht existierender String kann auch nicht umgewandelt werden!
     */
    if ( From->addr == NULL ) {
        XtDisplayStringConversionWarning(Dsp, From->addr, XmRXmString);
        return False;
    }
    /*
     * Wenn der Aufrufer Platz fuer den String reservierte, aber der nicht
     * ausreicht, dann Fehler an ihn zurueckmelden.
     */
    if ( (To->addr != NULL) && (To->size < sizeof(XmString)) ) {
        To->size = sizeof(XmString);
        return False;
    }
    /*
     * Ansonsten koennen wir denn String ruhigen Gewissens umwandeln. Je
     * nachdem, ob der Aufrufer Platz reservierte (fuer den opaquen Daten-
     * type XmString), geben wir den XmString direkt zurueck oder aber einen
     * Zeiger auf eine statische Variable, die den XmString haelt. Der Auf-
     * rufer muss in diesem Fall den erhaltenen Wert sofort umkopieren!
     */
    string = StringToXmString((char *) From->addr, NULL);
    if ( string == NULL ) {
        XtDisplayStringConversionWarning(Dsp, From->addr, XmRXmString);
        return False;
    }
    if ( To->addr )
        *(XmString *) To->addr = string;
    else
        To->addr = (XtPointer) &string;
    To->size = sizeof(XmString);
    return True;
} /* CvtStringToXmString */

/* --------------------------------------------------------------------------
 * Hier nun diejenige Prozedur, mittels derer der Konverter registriert wird.
 * Dabei wird der Konverter fuer alle Applikations-Kontexte aktiv - egal ob
 * diese bereits eingerichtet wurde oder erst noch werden.
 */
void XmRegisterNewXmStringConverter(XtCacheType CacheType)
{
    XtSetTypeConverter(XmRString, XmRXmString, CvtStringToXmString,
                       NULL, 0, CacheType, (XtDestructor) CleanupXmString);
} /* XmRegisterNewXmStringConverter */

/* Ende von String2XmString.c */
